// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'package:supabase_flutter/supabase_flutter.dart';
import 'package:wyatt_crud_bloc/wyatt_crud_bloc.dart' as crud show Query;
import 'package:wyatt_crud_bloc/wyatt_crud_bloc.dart' hide Query;

typedef QueryBuilder = PostgrestTransformBuilder<List<Map<String, dynamic>>>;
typedef StreamBuilder = SupabaseStreamBuilder;

/// {@template crud_data_source_supabase_impl}
/// A concrete implementation of [CrudDataSource] that uses
/// [Supabase] as the data source.
/// {@endtemplate}
class CrudDataSourceSupabaseImpl extends CrudDataSource {
  /// {@macro crud_data_source_supabase_impl}
  CrudDataSourceSupabaseImpl(
    String table, {
    required super.id,
    this.getIdKey,
    SupabaseClient? supabase,
  }) : _supabase = supabase ?? Supabase.instance.client {
    _supabaseQueryBuilder = _supabase.from(table);
  }

  final SupabaseClient _supabase;
  late SupabaseQueryBuilder _supabaseQueryBuilder;

  /// Get the id field name depending on the operation type.
  /// If null, 'id' will be used.
  final String? Function(
    OperationType operationType,
  )? getIdKey;

  @override
  Future<void> create(Map<String, dynamic> object, {String? id}) {
    if (id != null) {
      return _supabaseQueryBuilder.upsert(object);
    } else {
      final String? id = this.id(
        OperationType.create,
        object,
      );

      if (id != null) {
        return _supabaseQueryBuilder.upsert(object);
      }
      return _supabaseQueryBuilder.insert(object);
    }
  }

  @override
  Future<void> delete(String id) {
    final String idKey = getIdKey?.call(OperationType.delete) ?? 'id';
    return _supabaseQueryBuilder.delete().match({idKey: id});
  }

  @override
  Future<void> deleteAll() async {
    final List<Map<String, dynamic>> snapshots =
        await _supabaseQueryBuilder.select();
    for (final Map<String, dynamic> snapshot in snapshots) {
      final String idKey = getIdKey?.call(OperationType.delete) ?? 'id';
      await _supabaseQueryBuilder.delete().match({idKey: snapshot[idKey]});
    }
  }

  @override
  Future<Map<String, dynamic>?> get(String id) async {
    final String idKey = getIdKey?.call(OperationType.read) ?? 'id';
    final List<Map<String, dynamic>> snapshots =
        await _supabaseQueryBuilder.select().match({idKey: id});
    return snapshots.firstOrNull;
  }

  @override
  Future<List<Map<String, dynamic>?>> getAll() async =>
      _supabaseQueryBuilder.select();

  @override
  Future<List<Map<String, dynamic>?>> search(
    List<crud.Query> conditions,
  ) async {
    final PostgrestFilterBuilder<List<Map<String, dynamic>>> query =
        _supabaseQueryBuilder.select();

    final filter = conditions.whereType<WhereQuery<dynamic>>().firstOrNull;
    final modifiers = [
      ...conditions.whereType<LimitQuery>(),
      ...conditions.whereType<OrderByQuery>(),
    ];

    PostgrestTransformBuilder<List<Map<String, dynamic>>>? queryTransform;

    if (filter != null) {
      queryTransform = _queryFilterParser(filter, query);
    }

    for (final condition in modifiers) {
      queryTransform = _queryModifierParser(condition, queryTransform ?? query);
    }

    return query;
  }

  @override
  Stream<List<Map<String, dynamic>?>> stream({
    String? id,
    List<crud.Query>? conditions,
    bool includeMetadataChanges = false,
  }) {
    // TODO(hpcl): use conditions!!
    final stream = _supabaseQueryBuilder
        .stream(primaryKey: [getIdKey?.call(OperationType.read) ?? 'id']);

    if (id != null) {
      return stream.eq(
        getIdKey?.call(OperationType.read) ?? 'id',
        id,
      );
    } else {
      return stream;
    }
  }

  @override
  Future<void> update(
    String id, {
    Map<String, dynamic>? object,
  }) async {
    if (object != null) {
      final String idKey = getIdKey?.call(OperationType.update) ?? 'id';
      return _supabaseQueryBuilder.update(object).match({idKey: id});
    } else {
      throw Exception('You must provide data to update');
    }
  }

  @override
  Future<void> updateAll(Map<String, Object?>? data) async {
    if (data == null) {
      throw Exception('You must provide data to update');
    }
    return _supabaseQueryBuilder.update(data);
  }

  QueryBuilder _queryFilterParser(
    crud.Query condition,
    PostgrestFilterBuilder<List<Map<String, dynamic>>> query,
  ) {
    if (condition is WhereQuery) {
      switch (condition.type) {
        case WhereQueryType.isEqualTo:
          return query.eq(condition.field, condition.value as Object);
        case WhereQueryType.isNotEqualTo:
          return query.neq(condition.field, condition.value as Object);
        case WhereQueryType.isLessThan:
          return query.lt(condition.field, condition.value as Object);
        case WhereQueryType.isLessThanOrEqualTo:
          return query.lte(condition.field, condition.value as Object);
        case WhereQueryType.isGreaterThan:
          return query.gt(condition.field, condition.value as Object);
        case WhereQueryType.isGreaterThanOrEqualTo:
          return query.gte(condition.field, condition.value as Object);
        case WhereQueryType.arrayContains:
          return query.contains(condition.field, condition.value as Object);
        case WhereQueryType.whereIn:
          return query.inFilter(
            condition.field,
            condition.value as List<Object>,
          );
        case WhereQueryType.whereNotIn:
          return query.not(
            condition.field,
            'in',
            condition.value as List<Object>,
          );
        case WhereQueryType.isNull:
          if (condition.value as bool) {
            return query.isFilter(condition.field, null);
          } else {
            return query.not(condition.field, 'is', null);
          }
        case WhereQueryType.arrayContainsAny:
          throw UnsupportedError(
            '${condition.type} is not supported by Supabase',
          );
      }
    }
    return query;
  }

  StreamBuilder _streamFilterParser(
    crud.Query condition,
    SupabaseStreamFilterBuilder query,
  ) {
    if (condition is WhereQuery) {
      switch (condition.type) {
        case WhereQueryType.isEqualTo:
          return query.eq(condition.field, condition.value as Object);
        case WhereQueryType.isNotEqualTo:
          return query.neq(condition.field, condition.value as Object);
        case WhereQueryType.isLessThan:
          return query.lt(condition.field, condition.value as Object);
        case WhereQueryType.isLessThanOrEqualTo:
          return query.lte(condition.field, condition.value as Object);
        case WhereQueryType.isGreaterThan:
          return query.gt(condition.field, condition.value as Object);
        case WhereQueryType.isGreaterThanOrEqualTo:
          return query.gte(condition.field, condition.value as Object);
        case WhereQueryType.whereIn:
          return query.inFilter(
            condition.field,
            condition.value as List<Object>,
          );
        case WhereQueryType.arrayContains:
        case WhereQueryType.arrayContainsAny:
        case WhereQueryType.whereNotIn:
        case WhereQueryType.isNull:
          throw UnsupportedError(
            '${condition.type} is not supported in stream by Supabase',
          );
      }
    }
    return query;
  }

  QueryBuilder _queryModifierParser(
    crud.Query condition,
    QueryBuilder query,
  ) {
    if (condition is LimitQuery) {
      return query.limit(condition.limit);
    } else if (condition is OrderByQuery) {
      return query.order(
        condition.field,
        ascending: condition.ascending,
      );
    }
    return query;
  }
}
